#! /usr/bin/python3
# Copyright 2021-present StarRocks, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from textwrap import dedent
import time

from sqlalchemy import Connection, exc, schema as sa_schema
from sqlalchemy.dialects.mysql.pymysql import MySQLDialect_pymysql
from sqlalchemy.dialects.mysql.base import MySQLDDLCompiler, MySQLTypeCompiler, MySQLCompiler, MySQLIdentifierPreparer, _DecodingRow
from sqlalchemy.sql import sqltypes
from sqlalchemy.util import topological
from sqlalchemy import util
from sqlalchemy import log
from sqlalchemy.engine import reflection
from sqlalchemy.dialects.mysql.types import TINYINT, SMALLINT, INTEGER, BIGINT, DECIMAL, DOUBLE, FLOAT, CHAR, VARCHAR, DATETIME
from sqlalchemy.dialects.mysql.json import JSON
from .datatype import (
    LARGEINT, HLL, BITMAP, PERCENTILE, ARRAY, MAP, STRUCT,
    DATE
)

from . import reflection as _reflection


##############################################################################################
## NOTES - INCOMPLETE/UNFINISHED
# There are a number of items in here marked as ToDo
# In terms of table creation, the Partition, Distribution and OrderBy clauses need to be addressed from table options
# Tests `test_has_index` and `test_has_index_schema` are failing, this is because the CREATE INDEX statement appears to work async
#  and only when it's finished does it appear in the table definition
# Other tests are failing, need to fix or figure out how to suppress
# Review some skipped test suite requirements
#
#
#
##############################################################################################

# starrocks supported data types
ischema_names = {
    # === Boolean ===
    "boolean": sqltypes.BOOLEAN,
    # === Integer ===
    "tinyint": TINYINT,
    "smallint": SMALLINT,
    "int": INTEGER,
    "bigint": BIGINT,
    "largeint": LARGEINT,
    # === Floating-point ===
    "float": FLOAT,
    "double": DOUBLE,
    # === Fixed-precision ===
    "decimal": DECIMAL,
    "decimal32": DECIMAL,
    "decimal64": DECIMAL,
    "decimal128": DECIMAL,
    # === String ===
    "varchar": VARCHAR,
    "char": CHAR,
    "json": JSON,
    # === Date and time ===
    "date": DATE,
    "datetime": DATETIME,
    "timestamp": sqltypes.DATETIME,
    # == binary ==
    "binary": sqltypes.BINARY,
    "varbinary": sqltypes.VARBINARY,
    # === Structural ===
    "array": ARRAY,
    "map": MAP,
    "struct": STRUCT,
    "hll": HLL,
    "percentile": PERCENTILE,
    "bitmap": BITMAP,
}


class StarRocksTypeCompiler(MySQLTypeCompiler):

    def visit_BOOLEAN(self, type_, **kw):
        return "BOOLEAN"

    def visit_FLOAT(self, type_, **kw):
        return "FLOAT"

    def visit_TINYINT(self, type_, **kw):
        return "TINYINT"

    def visit_SMALLINT(self, type_, **kw):
        return "SMALLINT"

    def visit_INTEGER(self, type_, **kw):
        return "INTEGER"

    def visit_BIGINT(self, type_, **kw):
        return "BIGINT"

    def visit_LARGEINT(self, type_, **kw):
        return "LARGEINT"

    def visit_ARRAY(self, type_, **kw):
        """Compiles the ARRAY type into the correct StarRocks syntax."""
        inner_type_sql = self.process(type_.item_type, **kw)
        return f"ARRAY<{inner_type_sql}>"

    def visit_MAP(self, type_, **kw):
        return "MAP<keytype,valuetype>"

    def visit_STRUCT(self, type_, **kw):
        return "STRUCT<name, type>"

    def visit_HLL(self, type_, **kw):
        return "HLL"

    def visit_BITMAP(self, type_, **kw):
        return "BITMAP"


class StarRocksSQLCompiler(MySQLCompiler):
    def visit_delete(self, delete_stmt, **kw):
        result = super().visit_delete(delete_stmt, **kw)
        compile_state = delete_stmt._compile_state_factory(
            delete_stmt, self, **kw
        )
        delete_stmt = compile_state.statement
        table = self.delete_table_clause(
            delete_stmt, delete_stmt.table, False
        )
        if not delete_stmt._where_criteria:
            return "TRUNCATE TABLE " + table
        return result

    def limit_clause(self, select, **kw):
        # StarRocks supports:
        #   LIMIT <limit>
        #   LIMIT <limit> OFFSET <offset>
        text = ""
        if select._limit_clause is not None:
            text += "\n LIMIT " + self.process(select._limit_clause, **kw)
        if select._offset_clause is not None:
            # if select._limit_clause is None:
            #     text += "\n LIMIT -1"
            text += " OFFSET " + self.process(select._offset_clause, **kw)
        return text


class StarRocksDDLCompiler(MySQLDDLCompiler):

    def visit_create_table(self, create, **kw):
        table = create.element
        preparer = self.preparer

        text = "\nCREATE "
        if table._prefixes:
            text += " ".join(table._prefixes) + " "

        text += "TABLE "
        if create.if_not_exists:
            text += "IF NOT EXISTS "

        text += preparer.format_table(table) + " "

        create_table_suffix = self.create_table_suffix(table)
        if create_table_suffix:
            text += create_table_suffix + " "

        text += "("

        separator = "\n"

        # if only one primary key, specify it along with the column
        first_pk = False
        for create_column in create.columns:
            column = create_column.element
            try:
                processed = self.process(
                    create_column, first_pk=column.primary_key and not first_pk
                )
                if processed is not None:
                    text += separator
                    separator = ", \n"
                    text += "\t" + processed
                if column.primary_key:
                    first_pk = True
            except exc.CompileError as ce:
                raise exc.CompileError(
                    "(in table '%s', column '%s'): %s"
                    % (table.description, column.name, ce.args[0])
                ) from ce

        # N.B. Primary Key is specified in post_create_table
        #  Indexes are created by SQLA after the creation of the table using CREATE INDEX
        # const = self.create_table_constraints(
        #     table,
        #     _include_foreign_key_constraints=create.include_foreign_key_constraints,  # noqa
        # )
        # if const:
        #     text += separator + "\t" + const

        text += "\n)%s\n\n" % self.post_create_table(table)
        return text

    def post_create_table(self, table):
        """Build table-level CREATE options like ENGINE and COLLATE."""

        table_opts = []

        opts = dict(
            (k[len(self.dialect.name) + 1 :].upper(), v)
            for k, v in table.kwargs.items()
            if k.startswith("%s_" % self.dialect.name)
        )

        if table.comment is not None:
            opts["COMMENT"] = table.comment


        if 'ENGINE' in opts:
            table_opts.append(f'ENGINE={opts["ENGINE"]}')

        if 'PRIMARY_KEY' in opts:
            table_opts.append(f'PRIMARY KEY({opts["PRIMARY_KEY"]})')

        if 'DISTRIBUTED_BY' in opts:
            table_opts.append(f'DISTRIBUTED BY HASH({opts["DISTRIBUTED_BY"]})')

        if "COMMENT" in opts:
            comment = self.sql_compiler.render_literal_value(
                opts["COMMENT"], sqltypes.String()
            )
            table_opts.append(f"COMMENT {comment}")

        # ToDo - Partition
        # ToDo - Distribution

        if "ORDER_BY" in opts:
            table_opts.append(f"ORDER BY ({opts['ORDER_BY']})")

        if "PROPERTIES" in opts:
            props = ",\n".join([f'\t"{k}"="{v}"' for k, v in opts["PROPERTIES"]])
            table_opts.append(f"PROPERTIES(\n{props}\n)")

        return " ".join(table_opts)

    def get_column_specification(self, column, **kw):
        """Builds column DDL."""

        colspec = [
            self.preparer.format_column(column),
            self.dialect.type_compiler.process(
                column.type, type_expression=column
            ),
        ]

        # ToDo: Support aggregation type
        #  agg_type: aggregation type.If not specified, this column is key column.If specified, it is value
        #  column.The aggregation types supported are as follows:
        #  SUM, MAX, MIN, REPLACE
        #  HLL_UNION(only for HLL type)
        #  BITMAP_UNION(only for BITMAP)
        #  REPLACE_IF_NOT_NULL

        # if column.computed is not None:
        #     colspec.append(self.process(column.computed))

        # is_timestamp = isinstance(
        #     column.type._unwrapped_dialect_impl(self.dialect),
        #     sqltypes.TIMESTAMP,
        # )

        if not column.nullable:
            colspec.append("NOT NULL")

        # see: https://docs.sqlalchemy.org/en/latest/dialects/mysql.html#mysql_timestamp_null  # noqa
        # elif column.nullable and is_timestamp:
        #     colspec.append("NULL")

        # Column comment is not supported in Starrocks
        # comment = column.comment
        # if comment is not None:
        #     literal = self.sql_compiler.render_literal_value(
        #         comment, sqltypes.String()
        #     )
        #     colspec.append("COMMENT " + literal)

        # ToDo >= version 3.0
        if (
            column.table is not None
            and column is column.table._autoincrement_column
            and (
                column.server_default is None
                or isinstance(column.server_default, sa_schema.Identity)
            )
            and not (
                self.dialect.supports_sequences
                and isinstance(column.default, sa_schema.Sequence)
                and not column.default.optional
            )
        ):
            colspec[1] = "BIGINT" # ToDo - remove this, find way to fix the test
            colspec.append("AUTO_INCREMENT")
        else:
            default = self.get_column_default_string(column)
            if default == "AUTO_INCREMENT":
                colspec[1] = "BIGINT"
                colspec.append("AUTO_INCREMENT")

            elif default is not None:
                colspec.append("DEFAULT " + default)
        return " ".join(colspec)

    def visit_computed_column(self, generated, **kw):
        #ToDo >= version 3.1
        text = "AS (%s)" % self.sql_compiler.process(
            generated.sqltext, include_table=False, literal_binds=True
        )
        return text

    def visit_primary_key_constraint(self, constraint, **kw):
        if len(constraint) == 0:
            return ""
        text = ""
        # if constraint.name is not None:
        #     formatted_name = self.preparer.format_constraint(constraint)
        #     if formatted_name is not None:
        #         text += "CONSTRAINT %s " % formatted_name
        text += "PRIMARY KEY "
        text += "(%s)" % ", ".join(
            self.preparer.quote(c.name)
            for c in (
                constraint.columns_autoinc_first
                if constraint._implicit_generated
                else constraint.columns
            )
        )
        text += self.define_constraint_deferrability(constraint)
        return text

    def visit_set_table_comment(self, create, **kw):
        return "ALTER TABLE %s COMMENT=%s" % (
            self.preparer.format_table(create.element),
            self.sql_compiler.render_literal_value(
                create.element.comment, sqltypes.String()
            ),
        )

    def visit_drop_table_comment(self, create, **kw):
        return "ALTER TABLE %s COMMENT=''" % (
            self.preparer.format_table(create.element)
        )


class StarRocksIdentifierPreparer(MySQLIdentifierPreparer):
    # reserved_words = RESERVED_WORDS
    pass


@log.class_logger
class StarRocksDialect(MySQLDialect_pymysql):
    name = "starrocks"
    # Caching
    # Warnings are generated by SQLAlchemy if this flag is not explicitly set
    # and tests are needed before being enabled
    supports_statement_cache = True
    supports_server_side_cursors = False
    supports_empty_insert = False

    ischema_names = ischema_names


    statement_compiler = StarRocksSQLCompiler
    ddl_compiler = StarRocksDDLCompiler
    type_compiler = StarRocksTypeCompiler
    preparer = StarRocksIdentifierPreparer

    def __init__(self, *args, **kw):
        super(StarRocksDialect, self).__init__(*args, **kw)

    def _get_server_version_info(self, connection):
        # get database server version info explicitly over the wire
        # to avoid proxy servers like MaxScale getting in the
        # way with their own values, see #4205
        dbapi_con = connection.connection
        cursor = dbapi_con.cursor()
        cursor.execute("SELECT CURRENT_VERSION()")
        val = cursor.fetchone()[0]
        cursor.close()
        if isinstance(val, bytes):
            val = val.decode()

        return self._parse_server_version(val)

    def _parse_server_version(self, val):
        server_version_info = tuple()
        m = re.match(r"(\d+)\.?(\d+)?(?:\.(\d+))?(?:\.\d+)?(?:[-\s])?(?P<commit>.*)?", val)
        if m is not None:
            server_version_info = tuple([int(x) for x in m.group(1, 2, 3) if x is not None])

        # setting it here to help w the test suite
        self.server_version_info = server_version_info
        return server_version_info

    @util.memoized_property
    def _tabledef_parser(self) -> _reflection.StarRocksTableDefinitionParser:
        """return the StarRocksTableDefinitionParser, generate if needed.

        The deferred creation ensures that the dialect has
        retrieved server version information first.

        """
        preparer = self.identifier_preparer
        return _reflection.StarRocksTableDefinitionParser(self, preparer)

    def _read_from_information_schema(
        self, connection: Connection, inf_sch_table: str, charset: str | None = None, **kwargs
    ):
        def escape_single_quote(s):
            return s.replace("'", "\\'")
        
        st = dedent(f"""
            SELECT * 
            FROM information_schema.{inf_sch_table} 
            WHERE {" AND ".join([f"{k} = '{escape_single_quote(v)}'" for k, v in kwargs.items()])}
        """)
        rp = None
        try:
            rp = connection.execution_options(
                skip_user_error_events=False
            ).exec_driver_sql(st)
        except exc.DBAPIError as e:
            if self._extract_error_code(e.orig) == 1146:
                raise exc.NoSuchTableError(f"information_schema.{inf_sch_table}") from e
            else:
                raise
        rows = [_DecodingRow(row, charset) for row in rp.mappings().fetchall()]
        if not rows:
            raise exc.NoSuchTableError(f"Empty response for query: '{st}'")
        return rows
    
    @reflection.cache
    def _setup_parser(self, connection: Connection, table_name: str, schema: str | None = None, **kw):
        charset = self._connection_charset
        parser = self._tabledef_parser

        if not schema:
            schema = connection.dialect.default_schema_name

        table_rows = self._read_from_information_schema(
            connection=connection,
            inf_sch_table="tables",
            charset=charset,
            table_schema=schema,
            table_name=table_name,
        )
        if len(table_rows) > 1:
            raise exc.InvalidRequestError(
                f"Multiple tables found with name {table_name} in schema {schema}"
            )

        table_config_rows = self._read_from_information_schema(
            connection=connection,
            inf_sch_table="tables_config",
            charset=charset,
            table_schema=schema,
            table_name=table_name,
        )
        if len(table_rows) > 1:
            raise exc.InvalidRequestError(
                f"Multiple tables found with name {table_name} in schema {schema}"
            )

        column_rows = self._read_from_information_schema(
            connection=connection,
            inf_sch_table="columns",
            charset=charset,
            table_schema=schema,
            table_name=table_name,
        )

        return parser.parse(table=table_rows[0], table_config=table_config_rows[0], columns=column_rows, charset=charset)

    def _show_table_indexes(
        self, connection, table, charset=None, full_name=None
    ):
        """Run SHOW INDEX FROM for a ``Table``."""

        if full_name is None:
            full_name = self.identifier_preparer.format_table(table)
        st = "SHOW INDEX FROM %s" % full_name

        rp = None
        try:
            rp = connection.execution_options(
                skip_user_error_events=True
            ).exec_driver_sql(st)
        except exc.DBAPIError as e:
            if self._extract_error_code(e.orig) == 1146:
                raise exc.NoSuchTableError(full_name) from e
            else:
                raise
        index_results = self._compat_fetchall(rp, charset=charset)
        return index_results

    @reflection.cache
    def get_indexes(self, connection, table_name, schema=None, **kw):

        parsed_state = self._parsed_state_or_create(
            connection, table_name, schema, **kw
        )

        indexes = []

        for spec in parsed_state.keys:
            dialect_options = {}
            unique = False
            flavor = spec["type"]
            if flavor == "PRIMARY":
                continue
            if flavor == "DUPLICATE":
                continue
            if flavor == "UNIQUE":
                unique = True
            elif flavor in ("FULLTEXT", "SPATIAL"):
                dialect_options["%s_prefix" % self.name] = flavor
            elif flavor is None:
                pass
            else:
                self.logger.info(
                    "Converting unknown KEY type %s to a plain KEY", flavor
                )
                pass

            if spec["parser"]:
                dialect_options["%s_with_parser" % (self.name)] = spec[
                    "parser"
                ]

            index_d = {}

            index_d["name"] = spec["name"]
            index_d["column_names"] = [s[0] for s in spec["columns"]]
            mysql_length = {
                s[0]: s[1] for s in spec["columns"] if s[1] is not None
            }
            if mysql_length:
                dialect_options["%s_length" % self.name] = mysql_length

            index_d["unique"] = unique
            if flavor:
                index_d["type"] = flavor

            if dialect_options:
                index_d["dialect_options"] = dialect_options

            indexes.append(index_d)
        return indexes

    def has_table(self, connection, table_name, schema=None, **kw):
        try:
            return super().has_table(connection, table_name, schema, **kw)
        except exc.DBAPIError as e:
            if self._extract_error_code(e.orig) in (5501, 5502):
                return False
            raise



#! /usr/bin/python3
# Copyright 2021-present StarRocks, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import json
import re
from typing import Any

from sqlalchemy.dialects.mysql.types import DATETIME
from sqlalchemy.dialects.mysql.types import TIME
from sqlalchemy.dialects.mysql.types import TIMESTAMP
from sqlalchemy.dialects.mysql.base import _DecodingRow
from sqlalchemy.dialects.mysql.reflection import _re_compile
from sqlalchemy import log
from sqlalchemy import types as sqltypes
from sqlalchemy import util


@dataclasses.dataclass(kw_only=True)
class ReflectedState(object):
    """Stores informations about table or view."""

    table_name: str | None = None
    columns: list[dict] = dataclasses.field(default_factory=list)
    table_options: dict[str, str] = dataclasses.field(default_factory=dict)
    keys: list[dict] = dataclasses.field(default_factory=list)
    fk_constraints: list[dict] = dataclasses.field(default_factory=list)
    ck_constraints: list[dict] = dataclasses.field(default_factory=list)


@log.class_logger
class StarRocksTableDefinitionParser(object):
    """Parses content of information_schema tables to get table information."""

    def __init__(self, dialect, preparer):
        self.dialect = dialect
        self.preparer = preparer
        self._re_csv_int = _re_compile(r"\d+")
        self._re_csv_str = _re_compile(r"\x27(?:\x27\x27|[^\x27])*\x27")

    def parse(self, table: _DecodingRow, table_config: _DecodingRow, columns: list[_DecodingRow], charset: str) -> ReflectedState:
        return ReflectedState(
            table_name=table["TABLE_NAME"],
            columns=[self._parse_column(column=column) for column in columns],
            table_options=self._parse_table_options(table=table, table_config=table_config, columns=columns),
            keys=[{
                "type": self._get_key_type(columns=columns),
                "columns": [(c, None, None) for c in self._get_key_columns(columns=columns)],
                "parser": None,
                "name": None,
            }]
        )

    def _parse_column_type(self, column: _DecodingRow) -> Any:
        """
        Parse column type from information_schema.columns table.
        It splits column type into type and arguments.
        After that it creates instance of column type.
        """
        pattern = r"^(?P<type>\w+)(?:\s*\((?P<args>.*?)\))?$"
        match = re.match(pattern, column["COLUMN_TYPE"])
        type_ = match.group("type")
        args = match.group("args")
        try:
            col_type = self.dialect.ischema_names[type_]
        except KeyError:
            util.warn(
                "Did not recognize type '%s' of column '%s'" % (type_, column["COLUMN_NAME"])
            )
            col_type = sqltypes.NullType

        # Column type positional arguments eg. varchar(32)
        if args is None or args == "":
            type_args = []
        elif args[0] == "'" and args[-1] == "'":
            type_args = self._re_csv_str.findall(args)
        else:
            type_args = [int(v) for v in self._re_csv_int.findall(args)]

        # Column type keyword options
        type_kw = {}

        if issubclass(col_type, (DATETIME, TIME, TIMESTAMP)):
            if type_args:
                type_kw["fsp"] = type_args.pop(0)

        if col_type.__name__ == "LARGEINT":
            type_instance = col_type()
        else:
            type_instance = col_type(*type_args, **type_kw)
        return type_instance

    def _parse_column(self, column: _DecodingRow) -> dict:
        """
        Parse column from information_schema.columns table.
        It returns dictionary with column informations expected by sqlalchemy.
        """
        return {
            "name": column["COLUMN_NAME"],
            "type": self._parse_column_type(column=column),
            "nullable": column["IS_NULLABLE"] == "YES",
            "default": column["COLUMN_DEFAULT"],
            "autoincrement": None,  # TODO: This is not specified
            "computed": {
                "sqltext": column["GENERATION_EXPRESSION"]
            },
            "comment": column["COLUMN_COMMENT"],
        }

    def _get_key_columns(self, columns: list[_DecodingRow]) -> list[str]:
        """
        Get list of key columns from information_schema.columns table.
        It returns list of column names that are part of key.
        """
        sorted_columns = sorted(columns, key=lambda col: col["ORDINAL_POSITION"])
        return [c["COLUMN_NAME"] for c in sorted_columns if c["COLUMN_KEY"]]
    
    def _get_key_type(self, columns: list[_DecodingRow]) -> str:
        """
        Get key type from information_schema.columns table.
        It returns string representation of key type.
        It assumes only one key type is present in columns.
        """
        key_map = {
            "PRI": "PRIMARY",
            "UNI": "UNIQUE",
            "DUP": "DUPLICATE",
            "AGG": "AGGREGATE",
        }
        key_types = list(set([c["COLUMN_KEY"] for c in columns if c["COLUMN_KEY"]]))
        if len(key_types) > 1:
            raise NotImplementedError(f"Multiple key types found: {key_types}")
        if not key_types:
            return ""
        return key_map[key_types[0]]
    
    def _get_key_desc(self, columns: list[_DecodingRow]) -> str:
        """
        Get key description from information_schema.columns table.
        It returns string representation of key description.
        """
        quoted_cols = [self.preparer.quote_identifier(col) for col in self._get_key_columns(columns=columns)]
        return f"{self._get_key_type(columns=columns)} KEY({', '.join(quoted_cols)})"
    
    def _get_distribution(self, table_config: _DecodingRow) -> str:
        """
        Get distribution from information_schema.tables table.
        It returns string representation of distribution option.
        """
        distribution_cols = table_config["DISTRIBUTE_KEY"]
        distribution_str = f'({distribution_cols})' if distribution_cols else ""
        return f'{table_config["DISTRIBUTE_TYPE"]}{distribution_str}'
    
    def _parse_table_options(self, table: _DecodingRow, table_config: _DecodingRow, columns: list[_DecodingRow]) -> dict:
        """
        Parse table options from information_schema.tables table.
        It returns dictionary with table options expected by sqlalchemy.
        """
        return {
            f"{self.dialect.name}_engine": table_config["TABLE_ENGINE"],
            f"{self.dialect.name}_key_desc": self._get_key_desc(columns),
            f"{self.dialect.name}_comment": table["TABLE_COMMENT"],
            f"{self.dialect.name}_partition_by": table_config["PARTITION_KEY"],
            f"{self.dialect.name}_distribution": self._get_distribution(table_config),
            f"{self.dialect.name}_order_by": table_config["SORT_KEY"],
            f"{self.dialect.name}_properties": tuple(json.loads(table_config["PROPERTIES"] or "{}").items()),
        }



#! /usr/bin/python3
# Copyright 2021-present StarRocks, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import decimal
from textwrap import dedent

from sqlalchemy.testing.suite import *

from sqlalchemy.testing.assertions import AssertsCompiledSQL
from sqlalchemy import VARCHAR, Table, Column, Integer, MetaData
from sqlalchemy import schema

from sqlalchemy.testing import fixtures
from sqlalchemy import testing, literal
from sqlalchemy.testing.assertions import eq_
from sqlalchemy.sql.sqltypes import Float
from sqlalchemy.sql import elements


class CompileTest(fixtures.TestBase, AssertsCompiledSQL):

    __only_on__ = "starrocks"

    def test_create_table_with_properties(self):
        m = MetaData()
        tbl = Table(
            'atable', m, Column("id", Integer),
            starrocks_properties=(
                ("storage_medium", "SSD"),
                ("storage_cooldown_time", "2015-06-04 00:00:00"),
            ))
        self.assert_compile(
            schema.CreateTable(tbl),
            "CREATE TABLE atable (id INTEGER)PROPERTIES(\"storage_medium\"=\"SSD\",\"storage_cooldown_time\"=\"2015-06-04 00:00:00\")")

    def test_create_primary_key_table(self):
        m = MetaData()
        tbl = Table(
            'btable', m, Column("id", Integer, primary_key=True),
            starrocks_primary_key="id",
            starrocks_distributed_by="id"
            )
        self.assert_compile(
            schema.CreateTable(tbl),
            "CREATE TABLE btable (id BIGINT NOT NULL AUTO_INCREMENT)PRIMARY KEY(id) DISTRIBUTED BY HASH(id)")


# Float test fixes below for "Data type of first column cannot be FLOAT" error given by starrocks
class _LiteralRoundTripFixture2(object):
    supports_whereclause = True

    @testing.fixture
    def literal_round_trip(self, metadata, connection):
        """test literal rendering"""

        # for literal, we test the literal render in an INSERT
        # into a typed column.  we can then SELECT it back as its
        # official type; ideally we'd be able to use CAST here
        # but MySQL in particular can't CAST fully

        def run(type_, input_, output, filter_=None):
            t = Table("t", metadata, Column("a", Integer), Column("x", type_))
            t.create(connection)

            for value in input_:
                ins = (
                    t.insert()
                    .values(x=literal(value, type_))
                    .compile(
                        dialect=testing.db.dialect,
                        compile_kwargs=dict(literal_binds=True),
                    )
                )
                connection.execute(ins)

            if self.supports_whereclause:
                stmt = t.select().where(t.c.x == literal(value))
            else:
                stmt = t.select()

            stmt = stmt.compile(
                dialect=testing.db.dialect,
                compile_kwargs=dict(literal_binds=True),
            )
            for row in connection.execute(stmt):
                value = row[0]
                if filter_ is not None:
                    value = filter_(value)
                assert value in output

        return run


class FloatTest(_LiteralRoundTripFixture2, fixtures.TestBase):
    __backend__ = True

    @testing.fixture
    def do_numeric_test(self, metadata, connection):
        @testing.emits_warning(
            r".*does \*not\* support Decimal objects natively"
        )
        def run(type_, input_, output, filter_=None, check_scale=False):
            t = Table("t", metadata, Column("a", Integer), Column("x", type_))
            t.create(connection)
            connection.execute(t.insert(), [{"x": x} for x in input_])

            result = {row[1] for row in connection.execute(t.select())}
            output = set(output)
            if filter_:
                result = set(filter_(x) for x in result)
                output = set(filter_(x) for x in output)
            eq_(result, output)
            if check_scale:
                eq_([str(x) for x in result], [str(x) for x in output])

        return run

    @testing.requires.floats_to_four_decimals
    def test_float_as_decimal(self, do_numeric_test):
        do_numeric_test(
            Float(precision=8, asdecimal=True),
            [15.7563, decimal.Decimal("15.7563"), None],
            [decimal.Decimal("15.7563"), None],
            filter_=lambda n: n is not None and round(n, 4) or None,
        )

    def test_float_as_float(self, do_numeric_test):
        do_numeric_test(
            Float(precision=8),
            [15.7563, decimal.Decimal("15.7563")],
            [15.7563],
            filter_=lambda n: n is not None and round(n, 5) or None,
        )

    @testing.requires.precision_generic_float_type
    def test_float_custom_scale(self, do_numeric_test):
        do_numeric_test(
            Float(None, decimal_return_scale=6, asdecimal=True),
            [15.756382, decimal.Decimal("15.756382")],
            [decimal.Decimal("15.756382")],
            check_scale=True,
        )

    def test_render_literal_float(self, literal_round_trip):
        literal_round_trip(
            Float(4),
            [15.7563, decimal.Decimal("15.7563")],
            [15.7563],
            filter_=lambda n: n is not None and round(n, 5) or None,
        )


class StarRocksReflectionTest(fixtures.TestBase):
    __only_on__ = "starrocks"

    # @testing.combinations(True, False, argnames="primary_key")
    def test_default_reflection(self, connection, metadata):
        tn = "test_starrocks_reflection"
        Table(
            tn,
            metadata,
            Column("id", Integer),
            Column("data", VARCHAR(10)),
        )
        metadata.create_all(connection)

        reflected_meta = MetaData()
        reflected_meta.reflect(connection, only=[tn])
        reflected_table = reflected_meta.tables[tn]
        
        eq_(reflected_table.name, tn)
        eq_(reflected_table.schema, None)
        eq_(len(reflected_table.columns), 2)
        eq_(reflected_table.comment, "")
        eq_(set(reflected_table.columns.keys()), {"id", "data"})
        assert isinstance(reflected_table.columns["id"].type, Integer)
        assert isinstance(reflected_table.columns["data"].type, VARCHAR)
        eq_(reflected_table.columns["data"].type.length, 10)
        for c in ["id", "data"]:
            eq_(reflected_table.columns[c].comment, "")
            eq_(reflected_table.columns[c].nullable, True)
            eq_(reflected_table.columns[c].default, None)
            eq_(reflected_table.columns[c].autoincrement, None)
        eq_(reflected_table.dialect_options["starrocks"]["comment"], "")
        eq_(reflected_table.dialect_options["starrocks"]["distribution"], "RANDOM")
        eq_(reflected_table.dialect_options["starrocks"]["engine"], "OLAP")
        eq_(reflected_table.dialect_options["starrocks"]["key_desc"], "DUPLICATE KEY(`id`, `data`)")
        eq_(reflected_table.dialect_options["starrocks"]["order_by"], "`id`, `data`")
        eq_(reflected_table.dialect_options["starrocks"]["partition_by"], "")
        assert ("compression", "LZ4") in reflected_table.dialect_options["starrocks"]["properties"]

    def test_comment_reflection(self, connection, metadata):
        tn = "test_starrocks_reflection_comment"

        connection.execute(text(f"DROP TABLE IF EXISTS {tn}"))
        create_table_sql = dedent(f"""
            CREATE TABLE {tn} (
                id INT COMMENT 'This is id column',
                data VARCHAR(10) COMMENT 'This is data column'
            )
            COMMENT 'This is a test table'
        """).strip()
        connection.execute(text(create_table_sql))

        reflected_meta = MetaData()
        reflected_meta.reflect(connection, only=[tn])
        reflected_table = reflected_meta.tables[tn]

        eq_(reflected_table.comment, "This is a test table")
        eq_(reflected_table.columns["id"].comment, "This is id column")
        eq_(reflected_table.columns["data"].comment, "This is data column")

    def test_view_comment_reflection(self, connection, metadata):
        vn = "test_starrocks_reflection_view_comment"

        connection.execute(text(f"DROP VIEW IF EXISTS {vn}"))
        create_table_sql = dedent(f"""
            CREATE VIEW {vn} (
                id COMMENT 'This is id column',
                data COMMENT 'This is data column'
            )
            COMMENT 'This is a test view'
            AS SELECT 1 AS id, 'data' AS data
        """).strip()
        connection.execute(text(create_table_sql))

        reflected_meta = MetaData()
        reflected_meta.reflect(connection, only=[vn], views=True)
        reflected_table = reflected_meta.tables[vn]

        eq_(reflected_table.comment, "This is a test view")
        eq_(reflected_table.columns["id"].comment, "This is id column")
        eq_(reflected_table.columns["data"].comment, "This is data column")

    @testing.combinations(["id"], ["id", "data"], argnames="key_columns")
    @testing.combinations("PRIMARY", "DUPLICATE", "UNIQUE", argnames="key_type")
    def test_key_and_distribution_reflection(self, connection, metadata, key_type, key_columns):
        tn = "test_starrocks_reflection_key"
        connection.execute(text(f"DROP TABLE IF EXISTS {tn}"))
        create_table_sql = dedent(f"""
            CREATE TABLE {tn} (
                id INT,
                data VARCHAR(10),
                something_else DATETIME
            )
            {key_type} KEY ({', '.join(key_columns)})
            DISTRIBUTED BY HASH(id)
        """).strip()
        connection.execute(text(create_table_sql))

        reflected_meta = MetaData()
        reflected_meta.reflect(connection, only=[tn])
        reflected_table = reflected_meta.tables[tn]

        if key_type == "PRIMARY":
            for c in key_columns:
                eq_(reflected_table.columns[c].nullable, False)
        eq_(reflected_table.dialect_options["starrocks"]["key_desc"], f"{key_type} KEY({', '.join([f'`{c}`' for c in key_columns])})")
        eq_(reflected_table.dialect_options["starrocks"]["distribution"], "HASH(`id`)")

    @testing.combinations(["id"], ["id", "data"], argnames="partition_by")
    def test_partition_by_columns_reflection(self, connection, metadata, partition_by):
        tn = "test_starrocks_reflection_partition_by_columns"
        connection.execute(text(f"DROP TABLE IF EXISTS {tn}"))
        create_table_sql = dedent(f"""
            CREATE TABLE {tn} (
                id INT,
                data VARCHAR(10),
                something_else DATETIME
            )
            PARTITION BY ({', '.join(partition_by)})
        """).strip()
        connection.execute(text(create_table_sql))

        reflected_meta = MetaData()
        reflected_meta.reflect(connection, only=[tn])
        reflected_table = reflected_meta.tables[tn]

        eq_(reflected_table.dialect_options["starrocks"]["partition_by"], ', '.join([f'`{c}`' for c in partition_by]))

    def test_expression_partitioning_reflection(self, connection, metadata):
        tn = "test_starrocks_reflection_expression_partitioning"
        connection.execute(text(f"DROP TABLE IF EXISTS {tn}"))
        create_table_sql = dedent(f"""
            CREATE TABLE {tn} (
                event_day DATETIME NOT NULL,
                data VARCHAR(10)
            )
            PARTITION BY date_trunc('month', event_day)
        """).strip()
        connection.execute(text(create_table_sql))

        reflected_meta = MetaData()
        reflected_meta.reflect(connection, only=[tn])
        reflected_table = reflected_meta.tables[tn]

        eq_(reflected_table.dialect_options["starrocks"]["partition_by"], "`event_day`")

    @testing.combinations(["id"], ["id", "data"], argnames="order_by")
    def test_order_by_reflection(self, connection, metadata, order_by):
        tn = "test_starrocks_reflection_order_by"
        connection.execute(text(f"DROP TABLE IF EXISTS {tn}"))
        create_table_sql = dedent(f"""
            CREATE TABLE {tn} (
                id INT,
                data VARCHAR(10),
                something_else DATETIME
            )
            ORDER BY ({', '.join(order_by)})
        """).strip()
        connection.execute(text(create_table_sql))

        reflected_meta = MetaData()
        reflected_meta.reflect(connection, only=[tn])
        reflected_table = reflected_meta.tables[tn]

        eq_(reflected_table.dialect_options["starrocks"]["order_by"], ', '.join([f'`{c}`' for c in order_by]))

    def test_nullable_column_reflection(self, connection, metadata):
        tn = "test_starrocks_reflection_nullable"
        connection.execute(text(f"DROP TABLE IF EXISTS {tn}"))
        create_table_sql = dedent(f"""
            CREATE TABLE {tn} (
                id INT NOT NULL,
                data VARCHAR(10) NULL,
                something_else DATETIME
            )
        """).strip()
        connection.execute(text(create_table_sql))

        reflected_meta = MetaData()
        reflected_meta.reflect(connection, only=[tn])
        reflected_table = reflected_meta.tables[tn]

        eq_(reflected_table.columns["id"].nullable, False)
        eq_(reflected_table.columns["data"].nullable, True)

    def test_generated_columns_reflection(self, connection, metadata):
        tn = "test_starrocks_reflection_generated"
        connection.execute(text(f"DROP TABLE IF EXISTS {tn}"))
        create_table_sql = dedent(f"""
            CREATE TABLE {tn} (
                id INT,
                data JSON,
                generated_column BIGINT AS (id + 1),
                generated_json TEXT AS get_json_string(`data`, '$.some_key') COMMENT 'generated from json'
            )
        """).strip()
        connection.execute(text(create_table_sql))

        reflected_meta = MetaData()
        reflected_meta.reflect(connection, only=[tn])
        reflected_table = reflected_meta.tables[tn]

        eq_(reflected_table.columns["generated_column"].computed.sqltext.text, "id + 1")
        eq_(reflected_table.columns["generated_json"].computed.sqltext.text, "get_json_string(`data`, '$.some_key')")



#! /usr/bin/python3
# Copyright 2021-present StarRocks, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Optional, List, Any, Type, Dict, Callable, Literal
from datetime import date

from sqlalchemy.engine import Dialect
from sqlalchemy.sql import sqltypes
from sqlalchemy.sql.type_api import TypeEngine
from sqlalchemy.dialects.mysql.types import TINYINT, SMALLINT, INTEGER, BIGINT, DECIMAL, DOUBLE, FLOAT, CHAR, VARCHAR, DATETIME
from sqlalchemy.dialects.mysql.json import JSON

logger = logging.getLogger(__name__)


class LARGEINT(sqltypes.Integer):
    __visit_name__ = "LARGEINT"


class DATE(sqltypes.DATE):
    __visit_name__ = "DATE"
    def literal_processor(self, dialect: Dialect) -> Callable[[date], str]:
        def process(value: date) -> str:
            return f"TO_DATE('{value}')"

        return process


class HLL(sqltypes.Numeric):
    __visit_name__ = "HLL"


class BITMAP(sqltypes.Numeric):
    __visit_name__ = "BITMAP"


class PERCENTILE(sqltypes.Numeric):
    __visit_name__ = "PERCENTILE"


class ARRAY(TypeEngine):
    __visit_name__ = "ARRAY"

    def __init__(self, item_type, **kwargs):
        self.item_type = item_type
        super().__init__(**kwargs)

    @property
    def python_type(self) -> Optional[Type[List[Any]]]:
        return list


class MAP(TypeEngine):
    __visit_name__ = "MAP"

    @property
    def python_type(self) -> Optional[Type[Dict[Any, Any]]]:
        return dict


class STRUCT(TypeEngine):
    __visit_name__ = "STRUCT"

    @property
    def python_type(self) -> Optional[Type[Any]]:
        return None



#! /usr/bin/python3
# Copyright 2021-present StarRocks, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import re

from setuptools import find_packages, setup

_version_re = re.compile(r"__version__\s+=\s+(.*)")

with open("starrocks/__init__.py", "rb") as f:
    python_client_version = _version_re.search(f.read().decode("utf-8"))
    assert python_client_version is not None
    version = str(ast.literal_eval(python_client_version.group(1)))

with open('README.md') as readme:
    long_description = readme.read()

setup(
    name="starrocks",
    version=version,
    description="Python SQLAlchemy Dialect for StarRocks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache 2.0",
    author="StarRocks Team",
    url="https://github.com/StarRocks/starrocks",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Database :: Front-Ends",
    ],
    install_requires=[
        "sqlalchemy>=1.4",
        "sqlalchemy-utils>=0.41.2",
        "pymysql>=1.1.0",
        "alembic>=1.4.0"
    ],
    setup_requires=["pytest-runner"],
    tests_require=["pytest", "mock"],
    test_suite="test.test_suite",
    packages=find_packages(include=["starrocks"]),
    package_data={"": ["LICENSE", "README.md"]},
    include_package_data=True,
    zip_safe=False,
    entry_points={
        "sqlalchemy.dialects": [
            "starrocks = starrocks.dialect:StarRocksDialect",
            "starrocks.pymysql = starrocks.dialect:StarRocksDialect",
        ]
    },
)



from alembic.ddl.mysql import MySQLImpl
from sqlalchemy import Column
from sqlalchemy import BIGINT
from sqlalchemy import MetaData
from sqlalchemy import Table
from sqlalchemy import VARCHAR
from typing import Optional


class StarrocksImpl(MySQLImpl):
    __dialect__ = "starrocks"

    def version_table_impl(
        self,
        *,
        version_table: str,
        version_table_schema: Optional[str],
        version_table_pk: bool, # ignored as StarRocks requires a primary key
        **kw,
    ) -> Table:
        return Table(
            version_table,
            MetaData(),
            Column("id", BIGINT, autoincrement=True, primary_key=True),
            Column("version_num", VARCHAR(32), primary_key=False),
            schema=version_table_schema,
            starrocks_primary_key="id",
        )




#! /usr/bin/python3
# Copyright 2021-present StarRocks, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sqlalchemy.testing.requirements import SuiteRequirements
from sqlalchemy.testing import exclusions


class Requirements(SuiteRequirements):
    @property
    def unbounded_varchar(self):
        """Target database must support VARCHAR with no length"""
        return exclusions.closed()  # Starrocks does not support unbounded VARCHAR

    @property
    def temp_table_reflection(self):
        return exclusions.closed()

    @property
    def temp_table_reflect_indexes(self):
        return self.temp_table_reflection

    @property
    def temp_table_names(self):
        """target dialect supports listing of temporary table names"""
        return exclusions.closed()

    @property
    def has_temp_table(self):
        """target dialect supports checking a single temp table name"""
        return exclusions.closed()

    @property
    def temporary_tables(self):
        """target database supports temporary tables"""
        return exclusions.closed()

    @property
    def temporary_views(self):
        """target database supports temporary views"""
        return exclusions.closed()

    @property
    def index_reflection(self):
        # ToDo - open when multi-column index is supported
        return exclusions.closed()

    @property
    def primary_key_constraint_reflection(self):
        # ToDo - open when multi-column index is supported
        return exclusions.closed()

    @property
    def foreign_key_constraint_reflection(self):
        return exclusions.closed()

    @property
    def unique_constraint_reflection(self):
        """target dialect supports reflection of unique constraints"""
        return exclusions.closed()

    @property
    def binary_literals(self):
        """target backend supports simple binary literals, e.g. an
        expression like::

            SELECT CAST('foo' AS BINARY)

        Where ``BINARY`` is the type emitted from :class:`.LargeBinary`,
        e.g. it could be ``BLOB`` or similar.

        Basically fails on Oracle.

        """

        return exclusions.closed()

    @property
    def datetime_implicit_bound(self):
        """target dialect when given a datetime object will bind it such
        that the database server knows the object is a datetime, and not
        a plain string.
        https://github.com/sqlalchemy/sqlalchemy/discussions/9661
        """
        return exclusions.closed()

    @property
    def duplicate_key_raises_integrity_error(self):
        """target dialect raises IntegrityError when reporting an INSERT
        with a primary key violation.  (hint: it should)

        """
        return exclusions.closed()

    @property
    def offset(self):
        """target database can render OFFSET, or an equivalent, in a
        SELECT.
        """
        # ToDo - enable if Starrocks supports offset without limit (has order by)
        return exclusions.closed()

    @property
    def bound_limit_offset(self):
        """target database can render LIMIT and/or OFFSET using a bound
        parameter
        """
        # ToDo - see offset above
        return exclusions.closed()

    @property
    def sql_expression_limit_offset(self):
        """target database can render LIMIT and/or OFFSET with a complete
        SQL expression, such as one that uses the addition operator.
        parameter
        """

        return exclusions.closed()  # Not supported on StarRocks

    @property
    def json_type(self):
        """target platform implements a native JSON type."""

        return exclusions.closed()  # Starrocks returns all json elements casted as string

    @property
    def time(self):
        """target dialect supports representation of Python
        datetime.time() with microsecond objects."""

        return exclusions.closed()  # Not supported on Starrocks (no time type)    
    
    @property
    def time_microseconds(self):
        """target dialect supports representation of Python
        datetime.time() with microsecond objects."""

        return exclusions.closed()  # Not supported on Starrocks (no time type)
    
    @property
    def implements_get_lastrowid(self):
        """target dialect implements the executioncontext.get_lastrowid()
        method without reliance on RETURNING.
        """

        return exclusions.closed()
    
    @property
    def reflect_table_options(self):
        """Target database must support reflecting table_options."""
        
        return exclusions.open()
    
    @property
    def comment_reflection(self):
        """Indicates if the database support table comment reflection"""
        return exclusions.open()
    
    @property
    def sane_rowcount(self):
        return exclusions.closed()  # TODO: Check if that is expected
    
    @property
    def views(self):
        """Target database must support VIEWs."""

        return exclusions.open()



# ===================================================================================
# Below is the section with manual exclusions which cannot be excluded by Requirements
# ===================================================================================
from sqlalchemy.testing.suite.test_insert import InsertBehaviorTest
from sqlalchemy.testing.suite.test_dialect import ExceptionTest
from sqlalchemy.testing.suite.test_select import FetchLimitOffsetTest, LikeFunctionsTest
from sqlalchemy.testing.suite.test_ddl import LongNameBlowoutTest
from sqlalchemy.testing.suite.test_types import (
    DateTest, 
    DateTimeCoercedToDateTimeTest, 
    DateTimeTest, 
    JSONTest, 
    NumericTest, 
    StringTest,
    BinaryTest,
    EnumTest,
)
from sqlalchemy.testing.suite.test_reflection import (
    BizarroCharacterFKResolutionTest, 
    ComponentReflectionTest, 
    CompositeKeyReflectionTest,
    HasIndexTest,
    HasTableTest,
    QuotedNameArgumentTest,
)

# ========== Add missing requires. TODO: Can be deleted when https://github.com/sqlalchemy/sqlalchemy/pull/12362 is merged
BinaryTest.__requires__ = ("binary_literals",)
BizarroCharacterFKResolutionTest.__requires__ = ("primary_key_constraint_reflection",)
QuotedNameArgumentTest.test_get_foreign_keys = lambda *args: None  # missing requires.foreign_key_constraint_reflection
HasIndexTest.__requires__ = ("index_reflection",)
# Starrocks does not support FLOAT type for first column
NumericTest.test_float_as_decimal = lambda *args: None
NumericTest.test_float_as_float = lambda *args: None
NumericTest.test_float_custom_scale = lambda *args: None
NumericTest.test_render_literal_float = lambda *args: None
# Starrocks has no JSON_EXTRACT function
JSONTest.test_index_typed_access = lambda *args: None
JSONTest.test_index_typed_comparison = lambda *args: None
JSONTest.test_path_typed_comparison = lambda *args: None
# Syntax error -> Starrocks has no LIKE + ESCAPE
LikeFunctionsTest.test_contains_autoescape = lambda *args: None
LikeFunctionsTest.test_contains_autoescape_escape = lambda *args: None
LikeFunctionsTest.test_contains_escape = lambda *args: None
LikeFunctionsTest.test_endswith_autoescape = lambda *args: None
LikeFunctionsTest.test_endswith_autoescape_escape = lambda *args: None
LikeFunctionsTest.test_endswith_escape = lambda *args: None
LikeFunctionsTest.test_startswith_autoescape = lambda *args: None
LikeFunctionsTest.test_startswith_autoescape_escape = lambda *args: None
LikeFunctionsTest.test_startswith_escape = lambda *args: None
# Missing index_reflection
QuotedNameArgumentTest.test_get_indexes = lambda *args: None
# ======================================================
# ========== Not working "requires" decorators - they seems to be correctly used, but tests are not skipped
QuotedNameArgumentTest.test_get_unique_constraints = lambda *args: None
ComponentReflectionTest.test_get_multi_indexes = lambda *args: None
ComponentReflectionTest.test_get_multi_foreign_keys = lambda *args: None
ComponentReflectionTest.test_get_foreign_keys = lambda *args: None
ComponentReflectionTest.test_get_indexes = lambda *args: None
ComponentReflectionTest.test_get_multi_pk_constraint = lambda *args: None
ComponentReflectionTest.test_get_multi_unique_constraints = lambda *args: None
ComponentReflectionTest.test_get_noncol_index = lambda *args: None
ComponentReflectionTest.test_get_pk_constraint = lambda *args: None
ComponentReflectionTest.test_get_table_names = lambda *args: None
ComponentReflectionTest.test_get_temp_table_columns = lambda *args: None
ComponentReflectionTest.test_get_temp_table_indexes = lambda *args: None
ComponentReflectionTest.test_get_temp_table_unique_constraints = lambda *args: None
ComponentReflectionTest.test_get_unique_constraints = lambda *args: None
ComponentReflectionTest.test_get_unique_constraints = lambda *args: None
ComponentReflectionTest.test_reflect_table_temp_table = lambda *args: None
CompositeKeyReflectionTest.test_fk_column_order = lambda *args: None
CompositeKeyReflectionTest.test_pk_column_order = lambda *args: None
ExceptionTest.test_integrity_error = lambda *args: None
StringTest.test_nolength_string = lambda *args: None
FetchLimitOffsetTest.test_bound_offset = lambda *args: None
FetchLimitOffsetTest.test_bound_limit_offset = lambda *args: None
FetchLimitOffsetTest.test_expr_limit = lambda *args: None
FetchLimitOffsetTest.test_expr_limit_offset = lambda *args: None
FetchLimitOffsetTest.test_expr_limit_simple_offset = lambda *args: None
FetchLimitOffsetTest.test_expr_offset = lambda *args: None
FetchLimitOffsetTest.test_simple_limit_expr_offset = lambda *args: None
FetchLimitOffsetTest.test_simple_offset = lambda *args: None
FetchLimitOffsetTest.test_simple_offset_zero = lambda *args: None
LongNameBlowoutTest.test_long_convention_name = lambda *args: None
# ======================================================
# ========== Not implemented in reflection
ComponentReflectionTest.test_autoincrement_col = lambda *args: None  # There is no information about autoincrement in information_schema.columns
# Temporary table is only in informatio_schema.tables_config, but not in information_schema.tables and not in information_schema.columns
# ======================================================
# ========== Something is not working, but not in starrocks reflection or before checking reflection
ComponentReflectionTest.test_get_multi_columns = lambda *args: None  # Incorrect table created (e.g. DUP instead of PRI and no comments)
ComponentReflectionTest.test_get_view_names = lambda *args: None  # Views has not been created
DateTest.test_select_direct = lambda *args: None  # Compares string with date object
DateTimeCoercedToDateTimeTest.test_select_direct = lambda *args: None  # Compares string with datetime object
DateTimeTest.test_select_direct = lambda *args: None  # Compares string with datetime object
HasTableTest.test_has_table_cache = lambda *args: None  # clear_cache() is not needed?
InsertBehaviorTest.test_empty_insert = lambda *args: None  # Cannot "INSERT INTO autoinc_pk () VALUES ()""
EnumTest.__requires__ = ("binary_literals",)  # Fix Enum handling. Mysql has native ENUM type, but Starrocks has not
# ======================================================
